/*
 * Copyright (C) 2002 Red Hat, Inc.
 *
 * This is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ident "$Id: ring.c,v 1.14 2002/09/05 19:37:53 nalin Exp $"
#include "../config.h"
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include "debug.h"
#include "ring.h"

#ifdef VTE_DEBUG
static void
_vte_ring_validate(VteRing *ring)
{
	long i, max;
	max = ring->delta + ring->length;
	for (i = ring->delta; i < max; i++) {
		g_assert(_vte_ring_contains(ring, i));
		g_assert(ring->array[i % ring->max] != NULL);
	}
}
#endif

/**
 * _vte_ring_new:
 * @max_elements: the maximum size the new ring will be allowed to reach
 * @free: a #VteRingFreeFunc
 * @data: user data for @free
 *
 * Allocates a new ring capable of holding up to @max_elements elements at a
 * time, using @free to free them when they are removed from the ring.  The
 * @data pointer is passed to the @free callback whenever it is called.
 *
 * Returns: a new ring
 */
VteRing *
_vte_ring_new(glong max_elements, VteRingFreeFunc free, gpointer data)
{
	VteRing *ret = g_malloc0(sizeof(VteRing));
	ret->user_data = data;
	ret->delta = ret->length = 0;
	ret->max = MAX(max_elements, 2);
	ret->array = g_malloc0(sizeof(gpointer) * ret->max);
	ret->free = free;
	return ret;
}

VteRing *
_vte_ring_new_with_delta(glong max_elements, glong delta,
			VteRingFreeFunc free, gpointer data)
{
	VteRing *ret;
	ret = _vte_ring_new(max_elements, free, data);
	ret->delta = delta;
	return ret;
}

/**
 * _vte_ring_insert:
 * @ring: a #VteRing
 * @position: an index
 * @data: the new item
 *
 * Inserts a new item (@data) into @ring at the @position'th offset.  If @ring
 * already has an item stored at the desired location, it will be freed before
 * being replaced by the new @data.
 *
 */
void
_vte_ring_insert(VteRing *ring, long position, gpointer data)
{
	long point, i;
#ifdef VTE_DEBUG
	if (_vte_debug_on(VTE_DEBUG_RING)) {
		fprintf(stderr, "Inserting at position %ld.\n", position);
		fprintf(stderr, " Delta = %ld, Length = %ld, Max = %ld.\n",
			ring->delta, ring->length, ring->max);
	}
#endif
	g_return_if_fail(position >= ring->delta);
	g_return_if_fail(position <= ring->delta + ring->length);
	g_return_if_fail(data != NULL);

	/* Initial insertion, or append. */
	if (position == ring->length + ring->delta) {
		/* If there was something there before, free it. */
		if ((ring->free != NULL) &&
		    (ring->array[position % ring->max] != NULL)) {
#ifdef VTE_DEBUG
			if (_vte_debug_on(VTE_DEBUG_RING)) {
				fprintf(stderr, "Freeing item at position "
					"%ld.\n", position);
			}
#endif
			ring->free(ring->array[position % ring->max],
				   ring->user_data);
		}
		ring->array[position % ring->max] = data;
		if (ring->length == ring->max) {
			ring->delta++;
		} else {
			ring->length++;
		}
#ifdef VTE_DEBUG
		if (_vte_debug_on(VTE_DEBUG_RING)) {
			fprintf(stderr, " Delta = %ld, Length = %ld, "
				"Max = %ld.\n",
				ring->delta, ring->length, ring->max);
		}
		_vte_ring_validate(ring);
#endif
		return;
	}

	/* All other cases. */
	point = ring->delta + ring->length - 1;
	while (point <= 0) {
		point += ring->max;
	}

	/* If the buffer's full, then the last item will be lost. */
	if (ring->length == ring->max) {
		if (ring->free && ring->array[point % ring->max]) {
#ifdef VTE_DEBUG
			if (_vte_debug_on(VTE_DEBUG_RING)) {
				fprintf(stderr, "Freeing item at position "
					"%ld.\n", point);
			}
#endif
			ring->free(ring->array[point % ring->max],
				   ring->user_data);
		}
	}

	/* Bubble the rest down.  This isn't as slow as you probably think
	 * it is due to the pattern of usage. */
	if (ring->length != ring->max) {
		/* We'll need to copy the last item, too. */
		point++;
	}
	for (i = point; i > position; i--) {
		ring->array[i % ring->max] = ring->array[(i - 1) % ring->max];
	}
	ring->array[position % ring->max] = data;
	ring->length = MIN(ring->length + 1, ring->max);
#ifdef VTE_DEBUG
	if (_vte_debug_on(VTE_DEBUG_RING)) {
		fprintf(stderr, " Delta = %ld, Length = %ld, Max = %ld.\n",
			ring->delta, ring->length, ring->max);
	}
	_vte_ring_validate(ring);
#endif
}

/**
 * _vte_ring_remove:
 * @ring: a #VteRing
 * @position: an index
 * @free_element: TRUE if the item should be freed
 *
 * Removes the @position'th item from @ring, freeing it only if @free_element is
 * TRUE.
 *
 */
void
_vte_ring_remove(VteRing *ring, long position, gboolean free)
{
	long i;
#ifdef VTE_DEBUG
	if (_vte_debug_on(VTE_DEBUG_RING)) {
		fprintf(stderr, "Removing item at position %ld.\n", position);
		fprintf(stderr, " Delta = %ld, Length = %ld, Max = %ld.\n",
			ring->delta, ring->length, ring->max);
	}
#endif
	/* Remove the data at this position. */
	if (free && ring->array[position % ring->max] && ring->free) {
#ifdef VTE_DEBUG
		if (_vte_debug_on(VTE_DEBUG_RING)) {
			fprintf(stderr, "Freeing item at position %ld.\n",
				position);
		}
#endif
		ring->free(ring->array[position % ring->max], ring->user_data);
	}
	ring->array[position % ring->max] = NULL;

	/* Bubble the rest of the buffer up one notch.  This is also less
	 * of a problem than it might appear, again due to usage patterns. */
	for (i = position; i < ring->delta + ring->length - 1; i++) {
		ring->array[i % ring->max] = ring->array[(i + 1) % ring->max];
	}
	if (ring->length > 0) {
		ring->array[(ring->delta + ring->length - 1) % ring->max] = NULL;
		ring->length--;
	}
#ifdef VTE_DEBUG
	if (_vte_debug_on(VTE_DEBUG_RING)) {
		fprintf(stderr, " Delta = %ld, Length = %ld, Max = %ld.\n",
			ring->delta, ring->length, ring->max);
	}
	_vte_ring_validate(ring);
#endif
}

/**
 * _vte_ring_append:
 * @ring: a #VteRing
 * @data: the new item
 *
 * Appends a new item to the ring.  If an item must be removed to make room for
 * the new item, it is freed.
 *
 */
void
_vte_ring_append(VteRing *ring, gpointer data)
{
	_vte_ring_insert(ring, ring->delta + ring->length, data);
}

/**
 * _vte_ring_free:
 * @ring: a #VteRing
 * @free_elements: TRUE if items in the ring should be freed
 *
 * Frees the ring and, optionally, each of the items it contains.
 *
 */
void
_vte_ring_free(VteRing *ring, gboolean free_elements)
{
	long i;
	if (free_elements) {
		for (i = 0; i < ring->max; i++) {
			/* Remove this item. */
			if ((ring->free != NULL) && (ring->array[i] != NULL)) {
				ring->free(ring->array[i], ring->user_data);
				ring->array[i] = NULL;
			}
		}
	}
	g_free(ring->array);
	memset(ring, 0, sizeof(ring));
	g_free(ring);
}

#ifdef RING_MAIN
static void
scrolled_off(gpointer freed, gpointer data)
{
	long *l = (long *)freed;
	char *fmt = data;
	g_print(fmt, *l);
}

int
main(int argc, char **argv)
{
	long i, j, k, bias;
	const int size = 8;
	long values[24];
	long lone = 42;
	long *value;
	VteRing *ring;

	for (i = 0; i < G_N_ELEMENTS(values); i++) {
		values[i] = i;
	}

	ring = _vte_ring_new(size, scrolled_off, "Lost value %ld.\n");
	bias = 0;
	fprintf(stderr, "Initializing.\n");
	for (i = 0; i + bias <= G_N_ELEMENTS(values); i++) {
		k = 0;
		fprintf(stderr, "[%ld] ", i);
		for (j = 0; j < G_N_ELEMENTS(values); j++) {
			value = _vte_ring_index(ring, long *, j);
			if (value) {
				fprintf(stderr, "%s%ld->%ld",
					(k > 0) ? ", " : "",
					j, *value);
				k++;
			}
		}
		fprintf(stderr, "\n");
		fprintf(stderr, "[%ld] max %ld, delta %ld, length %ld = {",
			i, ring->max, ring->delta, ring->length);
		for (j = 0; j < size; j++) {
			value = ring->array[j];
			if (j > 0) {
				fprintf(stderr, ", ");
			}
			if (value) {
				fprintf(stderr, "%ld", *value);
			}
		}
		fprintf(stderr, "}\n");
		if (i == 3) {
			fprintf(stderr, "Removing item 3.\n");
			_vte_ring_remove(ring, 4, TRUE);
			bias--;
		} else
		if (i == 10) {
			fprintf(stderr, "Inserting item 7.\n");
			_vte_ring_insert(ring, 7, &lone);
			bias--;
		} else
		if (i == 20) {
			fprintf(stderr, "Inserting item 13.\n");
			_vte_ring_insert(ring, 13, &lone);
			bias--;
		} else
		if (i < G_N_ELEMENTS(values)) {
			fprintf(stderr, "Appending item.\n");
			_vte_ring_append(ring, &values[i + bias]);
		}
	}

	_vte_ring_free(ring, TRUE);

	return 0;
}
#endif
